#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2017 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c20671.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
"""

import numpy as np
from geotiepoints.geointerpolator import GeoInterpolator

lons = np.arange(20).reshape((4, 5), order="F")
lats = np.arange(20).reshape((4, 5), order="C")
lines = np.array([2, 7, 12, 17]) / 5.0
cols = np.array([2, 7, 12, 17, 22])
hlines = np.arange(20) / 5.0
hcols = np.arange(24)
satint = GeoInterpolator(
    (lons, lats), (lines, cols), (hlines, hcols), chunk_size=10)
satint.fill_borders('x', 'y')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2017 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c20671.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The tests package"""

from geotiepoints import interpolator

from geotiepoints.tests import (test_geointerpolator,
                                test_modis,
                                test_satelliteinterpolator,
                                test_interpolator,
                                test_multilinear,
                                test_modisinterpolator)


import sys
if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest

import doctest
import os
TRAVIS = os.environ.get("TRAVIS", False)


def suite():
    """The global test suite.
    """
    mysuite = unittest.TestSuite()
    if not TRAVIS:
        # Test sphinx documentation pages:
        mysuite.addTests(doctest.DocFileSuite('../../doc/source/index.rst'))
        # # Test the documentation strings
        mysuite.addTests(doctest.DocTestSuite(interpolator))

    # Use the unittests also
    mysuite.addTests(test_geointerpolator.suite())
    mysuite.addTests(test_modis.suite())
    mysuite.addTests(test_satelliteinterpolator.suite())
    mysuite.addTests(test_interpolator.suite())
    mysuite.addTests(test_multilinear.suite())
    mysuite.addTests(test_modisinterpolator.suite())
    return mysuite

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(suite())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2013 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Geographical interpolation (lon/lats).
"""

import numpy as np
from numpy import arccos, sign, rad2deg, sqrt, arcsin
from geotiepoints.interpolator import Interpolator

EARTH_RADIUS = 6370997.0

class GeoInterpolator(Interpolator):
    """
    Handles interpolation of geolocation from a grid of tie points.  It is
    preferable to have tie-points out till the edges if the tiepoint grid, but
    a method is provided to extrapolate linearly the tiepoints to the borders
    of the grid. The extrapolation is done automatically if it seems necessary.

    Uses numpy, scipy, and optionally pyresample

    The constructor takes in the tiepointed data as *data*, the
    *tiepoint_grid* and the desired *final_grid*. As optional arguments, one
    can provide *kx_* and *ky_* as interpolation orders (in x and y directions
    respectively), and the *chunksize* if the data has to be handled by pieces
    along the y axis (this affects how the extrapolator behaves). If
    *chunksize* is set, don't forget to adjust the interpolation orders
    accordingly: the interpolation is indeed done globaly (not chunkwise).
    """
    def __init__(self, lon_lat_data, *args, **kwargs):

        Interpolator.__init__(self, None, *args, **kwargs)
        self.lon_tiepoint = None
        self.lat_tiepoint = None
        try:
            # Maybe it's a pyresample object ?
            self.set_tiepoints(lon_lat_data.lons, lon_lat_data.lats)
            xyz = lon_lat_data.get_cartesian_coords()
            self.tie_data = [xyz[:, :, 0], xyz[:, :, 1], xyz[:, :, 2]]

        except AttributeError:
            self.set_tiepoints(lon_lat_data[0], lon_lat_data[1])
            lons_rad = np.radians(self.lon_tiepoint)
            lats_rad = np.radians(self.lat_tiepoint)
            x__ = EARTH_RADIUS * np.cos(lats_rad) * np.cos(lons_rad)
            y__ = EARTH_RADIUS * np.cos(lats_rad) * np.sin(lons_rad)
            z__ = EARTH_RADIUS * np.sin(lats_rad)
            self.tie_data = [x__, y__, z__]


        self.new_data = []
        for num in range(len(self.tie_data)):
            self.new_data.append([])


    def set_tiepoints(self, lon, lat):
        """Defines the lon,lat tie points.
        """
        self.lon_tiepoint = lon
        self.lat_tiepoint = lat

    def interpolate(self):
        newx, newy, newz = Interpolator.interpolate(self)
        lon = get_lons_from_cartesian(newx, newy)
        lat = get_lats_from_cartesian(newx, newy, newz)
        return lon, lat

def get_lons_from_cartesian(x__, y__):
    """Get longitudes from cartesian coordinates.
    """
    return rad2deg(arccos(x__ / sqrt(x__ ** 2 + y__ ** 2))) * sign(y__)
    
def get_lats_from_cartesian(x__, y__, z__, thr=0.8):
    """Get latitudes from cartesian coordinates.
    """
    # if we are at low latitudes - small z, then get the
    # latitudes only from z. If we are at high latitudes (close to the poles)
    # then derive the latitude using x and y:

    lats = np.where(np.logical_and(np.less(z__, thr * EARTH_RADIUS), 
                                   np.greater(z__, -1. * thr * EARTH_RADIUS)),
                    90 - rad2deg(arccos(z__/EARTH_RADIUS)),
                    sign(z__) *
                    (90 - rad2deg(arcsin(sqrt(x__ ** 2 + y__ ** 2)
                                         / EARTH_RADIUS))))
    return lats




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2013, 2017 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
"""

import unittest
import numpy as np

from geotiepoints.interpolator import Interpolator

TIES_EXP1 = np.array([[-2.00000000e+00,  -4.00000000e-01,   3.60000000e+00,
                       7.60000000e+00,   1.16000000e+01,   1.56000000e+01,
                       1.64000000e+01],
                      [-1.60000000e+00,   0.00000000e+00,   4.00000000e+00,
                       8.00000000e+00,   1.20000000e+01,   1.60000000e+01,
                       1.68000000e+01],
                      [-6.00000000e-01,   1.00000000e+00,   5.00000000e+00,
                       9.00000000e+00,   1.30000000e+01,   1.70000000e+01,
                       1.78000000e+01],
                      [-2.00000000e-01,   1.40000000e+00,   5.40000000e+00,
                       9.40000000e+00,   1.34000000e+01,   1.74000000e+01,
                       1.82000000e+01],
                      [4.44089210e-16,   1.60000000e+00,   5.60000000e+00,
                       9.60000000e+00,   1.36000000e+01,   1.76000000e+01,
                       1.84000000e+01],
                      [4.00000000e-01,   2.00000000e+00,   6.00000000e+00,
                       1.00000000e+01,   1.40000000e+01,   1.80000000e+01,
                       1.88000000e+01],
                      [1.40000000e+00,   3.00000000e+00,   7.00000000e+00,
                       1.10000000e+01,   1.50000000e+01,   1.90000000e+01,
                       1.98000000e+01],
                      [1.80000000e+00,   3.40000000e+00,   7.40000000e+00,
                       1.14000000e+01,   1.54000000e+01,   1.94000000e+01,
                       2.02000000e+01]])


class TestInterpolator(unittest.TestCase):

    # def test_fill_borders(self):
    #     lons = np.arange(20).reshape((4, 5), order="F")
    #     lats = np.arange(20).reshape((4, 5), order="C")
    #     lines = np.array([2, 7, 12, 17])
    #     cols = np.array([2, 7, 12, 17, 22])
    #     hlines = np.arange(20)
    #     hcols = np.arange(24)
    #     satint = Interpolator(
    #         (lons, lats), (lines, cols), (hlines, hcols), chunk_size=10)
    #     satint.fill_borders('x', 'y')
    #     self.assertTrue(np.allclose(satint.tie_data[0], TIES_EXP1))

    #     self.assertTrue(np.allclose(satint.row_indices,
    #                                 np.array([0,  2,  7,  9, 10, 12, 17, 19])))
    #     self.assertTrue(np.allclose(satint.col_indices,
    #                                 np.array([0,  2,  7, 12, 17, 22, 23])))

    def test_extrapolate_cols(self):
        lons = np.arange(10).reshape((2, 5), order="F")
        lats = np.arange(10).reshape((2, 5), order="C")
        lines = np.array([2, 7])
        cols = np.array([2, 7, 12, 17, 22])
        hlines = np.arange(10)
        hcols = np.arange(24)
        satint = Interpolator((lons, lats), (lines, cols), (hlines, hcols))
        self.assertTrue(np.allclose(satint._extrapolate_cols(satint.tie_data[0]),
                                    np.array([[-0.8,  0.,  2.,  4.,  6.,  8.,  8.4],
                                              [0.2,  1.,  3.,  5.,  7.,  9.,  9.4]])))

    def test_fill_col_borders(self):
        lons = np.arange(10).reshape((2, 5), order="F")
        lats = np.arange(10).reshape((2, 5), order="C")
        lines = np.array([2, 7])
        cols = np.array([2, 7, 12, 17, 22])
        hlines = np.arange(10)
        hcols = np.arange(24)
        satint = Interpolator((lons, lats), (lines, cols), (hlines, hcols))
        satint._fill_col_borders()
        self.assertTrue(np.allclose(satint.tie_data[0],
                                    np.array([[-0.8,  0.,  2.,  4.,  6.,  8.,  8.4],
                                              [0.2,  1.,  3.,  5.,  7.,  9.,  9.4]])))
        self.assertTrue(np.allclose(satint.col_indices,
                                    np.array([0,  2,  7, 12, 17, 22, 23])))

    def test_extrapolate_rows(self):
        lons = np.arange(10).reshape((2, 5), order="F")
        lats = np.arange(10).reshape((2, 5), order="C")
        lines = np.array([2, 7])
        cols = np.array([2, 7, 12, 17, 22])
        hlines = np.arange(10)
        hcols = np.arange(24)
        satint = Interpolator((lons, lats), (lines, cols), (hlines, hcols))
        first_idx = satint.hrow_indices[0]
        last_idx = satint.hrow_indices[-1]
        self.assertTrue(np.allclose(satint._extrapolate_rows(lons,
                                                             lines,
                                                             first_idx, last_idx),
                                    np.array([[-0.4,  1.6,  3.6,  5.6,  7.6],
                                              [0.,  2.,  4.,  6.,  8.],
                                              [1.,  3.,  5.,  7.,  9.],
                                              [1.4,  3.4,  5.4,  7.4,  9.4]])))

    def test_results_in_c_order(self):
        lons = np.arange(10).reshape((2, 5), order="F")
        lats = np.arange(10).reshape((2, 5), order="C")
        lines = np.array([2, 7])
        cols = np.array([2, 7, 12, 17, 22])
        hlines = np.arange(10)
        hcols = np.arange(24)
        satint = Interpolator((lons, lats), (lines, cols), (hlines, hcols))
        interp_results = satint.interpolate()
        assert interp_results[0].flags['C_CONTIGUOUS'] is True

    # def test_fill_row_borders(self):
    #     lons = np.arange(20).reshape((4, 5), order="F")
    #     lats = np.arange(20).reshape((4, 5), order="C")
    #     lines = np.array([2, 7, 12, 17])
    #     cols = np.array([2, 7, 12, 17, 22])
    #     hlines = np.arange(20)
    #     hcols = np.arange(24)
    #     satint = Interpolator((lons, lats), (lines, cols), (hlines, hcols))
    #     satint._fill_row_borders()
    #     self.assertTrue(np.allclose(satint.tie_data[0],
    #                                 np.array([[-0.4,   3.6,   7.6,  11.6,  15.6],
    #                                           [0.,   4.,   8.,  12.,  16.],
    #                                           [1.,   5.,   9.,  13.,  17.],
    #                                           [2.,   6.,  10.,  14.,  18.],
    #                                           [3.,   7.,  11.,  15.,  19.],
    #                                           [3.4,   7.4,  11.4,  15.4,  19.4]])))
    #     self.assertTrue(np.allclose(satint.row_indices,
    #                                 np.array([0,  2,  7, 12, 17, 19])))
    #     satint = Interpolator(
    #         (lons, lats), (lines, cols), (hlines, hcols), chunk_size=10)
    #     satint._fill_row_borders()
    #     self.assertTrue(np.allclose(satint.tie_data[0],
    #                                 np.array([[-0.4,   3.6,   7.6,  11.6,  15.6],
    #                                           [0.,   4.,   8.,  12.,  16.],
    #                                           [1.,   5.,   9.,  13.,  17.],
    #                                           [1.4,   5.4,   9.4,
    #                                               13.4,  17.4],
    #                                           [1.6,   5.6,   9.6,
    #                                               13.6,  17.6],
    #                                           [2.,   6.,  10.,  14.,  18.],
    #                                           [3.,   7.,  11.,  15.,  19.],
    #                                           [3.4,   7.4,  11.4,  15.4,  19.4]])))
    #     self.assertTrue(np.allclose(satint.row_indices,
    # np.array([0,  2,  7,  9, 10, 12, 17, 19])))


def suite():
    """The suite for Interpolator"""
    loader = unittest.TestLoader()
    mysuite = unittest.TestSuite()
    mysuite.addTest(loader.loadTestsFromTestCase(TestInterpolator))

    return mysuite



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2017 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c20671.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
"""

import h5py
import numpy as np
from geotiepoints import modis5kmto1km

FILENAME_FULL = 'testdata/test_5_to_1_geoloc_full.h5'
FILENAME_5KM = 'testdata/test_5_to_1_geoloc_5km.h5'

with h5py.File(FILENAME_FULL) as h5f:
    glons = h5f['longitude'][:] / 1000.
    glats = h5f['latitude'][:] / 1000.

with h5py.File(FILENAME_5KM) as h5f:
    lons = h5f['longitude'][:] / 1000.
    lats = h5f['latitude'][:] / 1000.

tlons, tlats = modis5kmto1km(lons, lats)

print np.allclose(tlons, glons, atol=0.05)
print np.allclose(tlats, glats, atol=0.05)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018 PyTroll community

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Interpolation of geographical tiepoints using the second order interpolation
scheme implemented in the CVIIRS software, as described here:
Compact VIIRS SDR Product Format User Guide (V1J)
http://www.eumetsat.int/website/wcm/idc/idcplg?IdcService=GET_FILE&dDocName=PDF_DMT_708025&RevisionSelectionMethod=LatestReleased&Rendition=Web
"""

import xarray as xr
import dask.array as da
import numpy as np

R = 6371.
# Aqua scan width and altitude in km
scan_width = 10.00017
H = 705.


def compute_phi(zeta):
    return np.arcsin(R * np.sin(zeta) / (R + H))


def compute_theta(zeta, phi):
    return zeta - phi


def compute_zeta(phi):
    return np.arcsin((R + H) * np.sin(phi) / R)


def compute_expansion_alignment(satz_a, satz_b, satz_c, satz_d):
    """All angles in radians."""
    zeta_a = satz_a
    zeta_b = satz_b

    phi_a = compute_phi(zeta_a)
    phi_b = compute_phi(zeta_b)
    theta_a = compute_theta(zeta_a, phi_a)
    theta_b = compute_theta(zeta_b, phi_b)
    phi = (phi_a + phi_b) / 2
    zeta = compute_zeta(phi)
    theta = compute_theta(zeta, phi)

    c_expansion = 4 * (((theta_a + theta_b) / 2 - theta) / (theta_a - theta_b))

    sin_beta_2 = scan_width / (2 * H)

    d = ((R + H) / R * np.cos(phi) - np.cos(zeta)) * sin_beta_2
    e = np.cos(zeta) - np.sqrt(np.cos(zeta) ** 2 - d ** 2)

    c_alignment = 4 * e * np.sin(zeta) / (theta_a - theta_b)

    return c_expansion, c_alignment


def get_corners(arr):
    arr_a = arr[:, :-1, :-1]
    arr_b = arr[:, :-1, 1:]
    arr_c = arr[:, 1:, 1:]
    arr_d = arr[:, 1:, :-1]
    return arr_a, arr_b, arr_c, arr_d


class ModisInterpolator():

    def __init__(self, cres, fres, cscan_full_width=None):
        if cres == 1000:
            self.cscan_len = 10
            self.cscan_width = 1
            self.cscan_full_width = 1354
        elif cres == 5000:
            self.cscan_len = 2
            self.cscan_width = 5
            if cscan_full_width is None:
                self.cscan_full_width = 271
            else:
                self.cscan_full_width = cscan_full_width

        if fres == 250:
            self.fscan_width = 4 * self.cscan_width
            self.fscan_full_width = 1354 * 4
            self.fscan_len = 4 * 10 // self.cscan_len
            self.get_coords = self._get_coords_1km
            self.expand_tiepoint_array = self._expand_tiepoint_array_1km
        elif fres == 500:
            self.fscan_width = 2 * self.cscan_width
            self.fscan_full_width = 1354 * 2
            self.fscan_len = 2 * 10 // self.cscan_len
            self.get_coords = self._get_coords_1km
            self.expand_tiepoint_array = self._expand_tiepoint_array_1km
        elif fres == 1000:
            self.fscan_width = 1 * self.cscan_width
            self.fscan_full_width = 1354
            self.fscan_len = 1 * 10 // self.cscan_len
            self.get_coords = self._get_coords_5km
            self.expand_tiepoint_array = self._expand_tiepoint_array_5km

    def _expand_tiepoint_array_1km(self, arr, lines, cols):
        arr = da.repeat(arr, lines, axis=1)
        arr = da.concatenate((arr[:, :lines//2, :], arr, arr[:, -(lines//2):, :]), axis=1)
        arr = da.repeat(arr.reshape((-1, self.cscan_full_width - 1)), cols, axis=1)
        return da.hstack((arr, arr[:, -cols:]))

    def _get_coords_1km(self, scans):
        y = (np.arange((self.cscan_len + 1) * self.fscan_len) % self.fscan_len) + .5
        y = y[self.fscan_len // 2:-(self.fscan_len // 2)]
        y[:self.fscan_len//2] = np.arange(-self.fscan_len/2 + .5, 0)
        y[-(self.fscan_len//2):] = np.arange(self.fscan_len + .5, self.fscan_len * 3 / 2)
        y = np.tile(y, scans)

        x = np.arange(self.fscan_full_width) % self.fscan_width
        x[-self.fscan_width:] = np.arange(self.fscan_width, self.fscan_width * 2)
        return x, y

    def _expand_tiepoint_array_5km(self, arr, lines, cols):
        arr = da.repeat(arr, lines * 2, axis=1)
        arr = da.repeat(arr.reshape((-1, self.cscan_full_width - 1)), cols, axis=1)
        if self.cscan_full_width == 271:
            return da.hstack((arr[:, :2], arr, arr[:, -2:]))
        else:
            return da.hstack((arr[:, :2], arr, arr[:, -5:], arr[:, -2:]))

    def _get_coords_5km(self, scans):
        y = np.arange(self.fscan_len * self.cscan_len) - 2
        y = np.tile(y, scans)

        x = (np.arange(self.fscan_full_width) - 2) % self.fscan_width
        x[0] = -2
        x[1] = -1
        if self.cscan_full_width == 271:
            x[-2] = 5
            x[-1] = 6
        elif self.cscan_full_width == 270:
            x[-7] = 5
            x[-6] = 6
            x[-5] = 7
            x[-4] = 8
            x[-3] = 9
            x[-2] = 10
            x[-1] = 11
        else:
            raise NotImplementedError("Can't interpolate if 5km tiepoints have less than 270 columns.")
        return x, y

    def interpolate(self, lon1, lat1, satz1):
        cscan_len = self.cscan_len
        cscan_full_width = self.cscan_full_width

        fscan_width = self.fscan_width
        fscan_len = self.fscan_len

        scans = satz1.shape[0] // cscan_len
        satz1 = satz1.data

        satz1 = satz1.reshape((-1, cscan_len, cscan_full_width))

        satz_a, satz_b, satz_c, satz_d = get_corners(da.deg2rad(satz1))

        c_exp, c_ali = compute_expansion_alignment(satz_a, satz_b, satz_c, satz_d)

        x, y = self.get_coords(scans)
        i_rs, i_rt = da.meshgrid(x, y)

        p_os = 0
        p_ot = 0

        s_s = (p_os + i_rs) * 1. / fscan_width
        s_t = (p_ot + i_rt) * 1. / fscan_len

        cols = fscan_width
        lines = fscan_len

        c_exp_full = self.expand_tiepoint_array(c_exp, lines, cols)
        c_ali_full = self.expand_tiepoint_array(c_ali, lines, cols)

        a_track = s_t
        a_scan = (s_s + s_s * (1 - s_s) * c_exp_full + s_t*(1 - s_t) * c_ali_full)

        res = []

        sublat = lat1[::16, ::16]
        sublon = lon1[::16, ::16]
        to_cart = abs(sublat).max() > 60 or (sublon.max() - sublon.min()) > 180

        if to_cart:
            datasets = lonlat2xyz(lon1, lat1)
        else:
            datasets = [lon1, lat1]

        for data in datasets:
            data_attrs = data.attrs
            dims = data.dims
            data = data.data
            data = data.reshape((-1, cscan_len, cscan_full_width))
            data_a, data_b, data_c, data_d = get_corners(data)
            data_a = self.expand_tiepoint_array(data_a, lines, cols)
            data_b = self.expand_tiepoint_array(data_b, lines, cols)
            data_c = self.expand_tiepoint_array(data_c, lines, cols)
            data_d = self.expand_tiepoint_array(data_d, lines, cols)

            data_1 = (1 - a_scan) * data_a + a_scan * data_b
            data_2 = (1 - a_scan) * data_d + a_scan * data_c
            data = (1 - a_track) * data_1 + a_track * data_2

            res.append(xr.DataArray(data, attrs=data_attrs, dims=dims))

        if to_cart:
            return xyz2lonlat(*res)
        else:
            return res


def modis_1km_to_250m(lon1, lat1, satz1):

    interp = ModisInterpolator(1000, 250)
    return interp.interpolate(lon1, lat1, satz1)


def modis_1km_to_500m(lon1, lat1, satz1):

    interp = ModisInterpolator(1000, 500)
    return interp.interpolate(lon1, lat1, satz1)


def modis_5km_to_1km(lon1, lat1, satz1):

    interp = ModisInterpolator(5000, 1000, lon1.shape[1])
    return interp.interpolate(lon1, lat1, satz1)

def lonlat2xyz(lons, lats):
    """Convert lons and lats to cartesian coordinates."""
    R = 6370997.0
    x_coords = R * da.cos(da.deg2rad(lats)) * da.cos(da.deg2rad(lons))
    y_coords = R * da.cos(da.deg2rad(lats)) * da.sin(da.deg2rad(lons))
    z_coords = R * da.sin(da.deg2rad(lats))
    return x_coords, y_coords, z_coords

def xyz2lonlat(x__, y__, z__):
    """Get longitudes from cartesian coordinates.
    """
    R = 6370997.0
    lons = da.rad2deg(da.arccos(x__ / da.sqrt(x__ ** 2 + y__ ** 2))) * da.sign(y__)
    lats = da.sign(z__) * (90 - da.rad2deg(da.arcsin(da.sqrt(x__ ** 2 + y__ ** 2) / R)))

    return lons, lats



"""
Unit tests for python-geotiepoints: MODIS examples
"""

import unittest
import numpy as np
import h5py
import os

FILENAME_250M_RESULT = os.path.join(
    os.path.dirname(__file__), '../../testdata/250m_lonlat_section_result.h5')
FILENAME_250M_INPUT = os.path.join(
    os.path.dirname(__file__), '../../testdata/250m_lonlat_section_input.h5')

FILENAME_FULL = os.path.join(
    os.path.dirname(__file__), '../../testdata/test_5_to_1_geoloc_full.h5')
FILENAME_5KM = os.path.join(
    os.path.dirname(__file__), '../../testdata/test_5_to_1_geoloc_5km.h5')

from geotiepoints import (modis5kmto1km, modis1kmto250m)

from geotiepoints import get_scene_splits


class TestUtils(unittest.TestCase):

    """Class for unit testing the ancillary interpolation functions
    """

    def setUp(self):
        pass

    def test_get_numof_subscene_lines(self):
        """Test getting the number of sub-scene lines dependent on the number
        of CPUs and for various number of lines in a scan"""

        ncpus = 3
        scene_splits = get_scene_splits(1060,
                                        10, ncpus)


class TestMODIS(unittest.TestCase):

    """Class for system testing the MODIS interpolation.
    """

    def setUp(self):
        pass

    def test_5_to_1(self):
        """test the 5km to 1km interpolation facility
        """

        with h5py.File(FILENAME_FULL) as h5f:
            glons = h5f['longitude'][:] / 1000.
            glats = h5f['latitude'][:] / 1000.

        with h5py.File(FILENAME_5KM) as h5f:
            lons = h5f['longitude'][:] / 1000.
            lats = h5f['latitude'][:] / 1000.

        tlons, tlats = modis5kmto1km(lons, lats)

        self.assert_(np.allclose(tlons, glons, atol=0.05))
        self.assert_(np.allclose(tlats, glats, atol=0.05))

    def test_1000m_to_250m(self):
        """test the 1 km to 250 meter interpolation facility
        """

        with h5py.File(FILENAME_250M_RESULT) as h5f:
            glons = h5f['longitude'][:] / 1000.
            glats = h5f['latitude'][:] / 1000.

        with h5py.File(FILENAME_250M_INPUT) as h5f:
            lons = h5f['longitude'][:] / 1000.
            lats = h5f['latitude'][:] / 1000.

        tlons, tlats = modis1kmto250m(lons, lats)

        self.assert_(np.allclose(tlons, glons, atol=0.05))
        self.assert_(np.allclose(tlats, glats, atol=0.05))

        tlons, tlats = modis1kmto250m(lons, lats, cores=4)

        self.assert_(np.allclose(tlons, glons, atol=0.05))
        self.assert_(np.allclose(tlats, glats, atol=0.05))


def suite():
    """The suite for MODIS"""
    loader = unittest.TestLoader()
    mysuite = unittest.TestSuite()
    mysuite.addTest(loader.loadTestsFromTestCase(TestUtils))
    mysuite.addTest(loader.loadTestsFromTestCase(TestMODIS))

    return mysuite

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018 Martin Raspaud

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy as np
import h5py
import os
from geotiepoints.modisinterpolator import modis_1km_to_250m, modis_1km_to_500m, modis_5km_to_1km
FILENAME_DATA = os.path.join(
    os.path.dirname(__file__), '../../testdata/modis_test_data.h5')

def to_da(arr):
    import xarray as xr
    import dask.array as da

    return xr.DataArray(da.from_array(arr, chunks=4096), dims=['y', 'x'])

class TestModisInterpolator(unittest.TestCase):
    def test_modis(self):
        h5f = h5py.File(FILENAME_DATA, 'r')
        lon1 = to_da(h5f['lon_1km'])
        lat1 = to_da(h5f['lat_1km'])
        satz1 = to_da(h5f['satz_1km'])

        lon250 = to_da(h5f['lon_250m'])
        lon500 = to_da(h5f['lon_500m'])

        lat250 = to_da(h5f['lat_250m'])
        lat500 = to_da(h5f['lat_500m'])

        lons, lats = modis_1km_to_250m(lon1, lat1, satz1)
        self.assertTrue(np.allclose(lon250, lons, atol=1e-2))
        self.assertTrue(np.allclose(lat250, lats, atol=1e-2))

        lons, lats = modis_1km_to_500m(lon1, lat1, satz1)
        self.assertTrue(np.allclose(lon500, lons, atol=1e-2))
        self.assertTrue(np.allclose(lat500, lats, atol=1e-2))

        lat5 = lat1[2::5, 2::5]
        lon5 = lon1[2::5, 2::5]

        satz5 = satz1[2::5, 2::5]
        lons, lats = modis_5km_to_1km(lon5, lat5, satz5)
        self.assertTrue(np.allclose(lon1, lons, atol=1e-2))
        self.assertTrue(np.allclose(lat1, lats, atol=1e-2))

        # Test level 2
        lat5 = lat1[2::5, 2:-5:5]
        lon5 = lon1[2::5, 2:-5:5]

        satz5 = satz1[2::5, 2:-5:5]
        lons, lats = modis_5km_to_1km(lon5, lat5, satz5)
        self.assertTrue(np.allclose(lon1, lons, atol=1e-2))
        self.assertTrue(np.allclose(lat1, lats, atol=1e-2))
        
    def test_poles_datum(self):
        import xarray as xr
        h5f = h5py.File(FILENAME_DATA, 'r')
        orig_lon = to_da(h5f['lon_1km'])
        lon1 = orig_lon + 180
        lon1 = xr.where(lon1 > 180, lon1 - 360, lon1)
        lat1 = to_da(h5f['lat_1km'])
        satz1 = to_da(h5f['satz_1km'])

        lat5 = lat1[2::5, 2::5]
        lon5 = lon1[2::5, 2::5]

        satz5 = satz1[2::5, 2::5]
        lons, lats = modis_5km_to_1km(lon5, lat5, satz5)
        lons = lons + 180
        lons = xr.where(lons > 180, lons - 360, lons)
        self.assertTrue(np.allclose(orig_lon, lons, atol=1e-2))
        self.assertTrue(np.allclose(lat1, lats, atol=1e-2))


def suite():
    """The suite for MODIS"""
    loader = unittest.TestLoader()
    mysuite = unittest.TestSuite()
    mysuite.addTest(loader.loadTestsFromTestCase(TestModisInterpolator))

    return mysuite

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2012-2018 PyTroll community

# Author(s):

#   Adam Dybbroe <adam.dybbroe@smhi.se>
#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Setting up the geo_interpolator project.
"""

import os
import sys

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext as _build_ext
import versioneer

requirements = ['numpy', 'scipy', 'pandas'],
# unittest2 is required by h5py 2.8.0rc:
test_requires = ['h5py', 'unittest2', 'xarray', 'dask']

if sys.platform.startswith("win"):
    extra_compile_args = []
else:
    extra_compile_args = ["-O3"]

EXTENSIONS = [
    Extension(
        'geotiepoints.multilinear_cython',
        sources=['geotiepoints/multilinear_cython.pyx'],
        extra_compile_args=extra_compile_args
    ),
    Extension(
        'geotiepoints.multilinear_cython',
        sources=['geotiepoints/multilinear_cython.pyx',
                 'geotiepoints/multilinear_cython.c'],
        language='c', extra_compile_args=extra_compile_args,
        depends=[]
    )

]

try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None


def set_builtin(name, value):
    if isinstance(__builtins__, dict):
        __builtins__[name] = value
    else:
        setattr(__builtins__, name, value)


cmdclass = versioneer.get_cmdclass()
versioneer_build_ext = cmdclass.get('build_ext', _build_ext)


class build_ext(versioneer_build_ext):

    """Work around to bootstrap numpy includes in to extensions.

    Copied from:

        http://stackoverflow.com/questions/19919905/how-to-bootstrap-numpy-installation-in-setup-py

    """

    def finalize_options(self):
        versioneer_build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        set_builtin('__NUMPY_SETUP__', False)
        import numpy
        self.include_dirs.append(numpy.get_include())


cmdclass['build_ext'] = build_ext

if __name__ == "__main__":
    if not os.getenv("USE_CYTHON", False) or cythonize is None:
        print(
            "Cython will not be used. Use environment variable 'USE_CYTHON=True' to use it")

        def cythonize(extensions, **dummy):
            """Fake function to compile from C/C++ files instead of compiling .pyx files with cython.
            """
            for extension in extensions:
                sources = []
                for sfile in extension.sources:
                    path, ext = os.path.splitext(sfile)
                    if ext in ('.pyx', '.py'):
                        if extension.language == 'c++':
                            ext = '.cpp'
                        else:
                            ext = '.c'
                        sfile = path + ext
                    sources.append(sfile)
                extension.sources[:] = sources
            return extensions

    setup(name='python-geotiepoints',
          version=versioneer.get_version(),
          description='Interpolation of geographic tiepoints in Python',
          author='Adam Dybbroe, Martin Raspaud',
          author_email='martin.raspaud@smhi.se',
          classifiers=["Development Status :: 4 - Beta",
                       "Intended Audience :: Science/Research",
                       "License :: OSI Approved :: GNU General Public License v3 " +
                       "or later (GPLv3+)",
                       "Operating System :: OS Independent",
                       "Programming Language :: Python",
                       "Topic :: Scientific/Engineering"],
          url="https://github.com/pytroll/python-geotiepoints",
          packages=['geotiepoints'],
          # packages=find_packages(),
          setup_requires=['numpy'],
          python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*',
          cmdclass=cmdclass,

          install_requires=requirements,
          ext_modules=cythonize(EXTENSIONS),

          test_suite='geotiepoints.tests.suite',
          tests_require=test_requires,
          zip_safe=False
          )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2013-2018 PyTroll community

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generic interpolation routines.
"""
import numpy as np
from scipy.interpolate import RectBivariateSpline, splev, splrep


def generic_modis5kmto1km(*data5km):
    """Getting 1km data for modis from 5km tiepoints.
    """
    cols5km = np.arange(2, 1354, 5)
    cols1km = np.arange(1354)
    lines = data5km[0].shape[0] * 5
    rows5km = np.arange(2, lines, 5)
    rows1km = np.arange(lines)

    along_track_order = 1
    cross_track_order = 3

    satint = Interpolator(list(data5km),
                          (rows5km, cols5km),
                          (rows1km, cols1km),
                          along_track_order,
                          cross_track_order,
                          chunk_size=10)
    satint.fill_borders("y", "x")
    return satint.interpolate()

# NOTE: extrapolate on a sphere ?


def _linear_extrapolate(pos, data, xev):
    """

    >>> import numpy as np
    >>> pos = np.array([1, 2])
    >>> data = np.arange(10).reshape((2, 5), order="F")
    >>> xev = 5
    >>> retv = _linear_extrapolate(pos, data, xev)
    >>> print([val for val in retv])
    [4.0, 6.0, 8.0, 10.0, 12.0]
    >>> xev = 0
    >>> retv = _linear_extrapolate(pos, data, xev)
    >>> print([val for val in retv])
    [-1.0, 1.0, 3.0, 5.0, 7.0]
    """
    if len(data) != 2 or len(pos) != 2:
        raise ValueError("len(pos) and the number of lines of data"
                         " must be 2.")

    return data[1] + ((xev - pos[1]) / (1.0 * (pos[0] - pos[1])) *
                      (data[0] - data[1]))


class Interpolator(object):

    """
    Handles interpolation of data from a grid of tie points.  It is preferable
    to have tie-points out till the edges if the tiepoint grid, but a method is
    provided to extrapolate linearly the tiepoints to the borders of the
    grid. The extrapolation is done automatically if it seems necessary.

    Uses numpy and scipy.

    The constructor takes in the tiepointed data as *data*, the
    *tiepoint_grid* and the desired *final_grid*. As optional arguments, one
    can provide *kx_* and *ky_* as interpolation orders (in x and y directions
    respectively), and the *chunksize* if the data has to be handled by pieces
    along the y axis (this affects how the extrapolator behaves). If
    *chunksize* is set, don't forget to adjust the interpolation orders
    accordingly: the interpolation is indeed done globaly (not chunkwise).
    """

    def __init__(self, data, tiepoint_grid, final_grid,
                 kx_=1, ky_=1, chunk_size=0):
        self.row_indices = tiepoint_grid[0]
        self.col_indices = tiepoint_grid[1]
        self.hrow_indices = final_grid[0]
        self.hcol_indices = final_grid[1]
        self.chunk_size = chunk_size
        if not isinstance(data, (tuple, list)):
            self.tie_data = [data]
        else:
            self.tie_data = list(data)

        self.new_data = []
        for num in range(len(self.tie_data)):
            self.new_data.append([])

        self.kx_, self.ky_ = kx_, ky_

    def fill_borders(self, *args):
        """Extrapolate tiepoint lons and lats to fill in the border of the
        chunks.
        """

        to_run = []
        cases = {"y": self._fill_row_borders,
                 "x": self._fill_col_borders}
        for dim in args:
            try:
                to_run.append(cases[dim])
            except KeyError:
                raise NameError("Unrecognized dimension: " + str(dim))

        for fun in to_run:
            fun()

    def _extrapolate_cols(self, data, first=True, last=True):
        """Extrapolate the column of data, to get the first and last together
        with the data.

        """

        if first:
            pos = self.col_indices[:2]
            first_column = _linear_extrapolate(pos,
                                               (data[:, 0], data[:, 1]),
                                               self.hcol_indices[0])
        if last:
            pos = self.col_indices[-2:]
            last_column = _linear_extrapolate(pos,
                                              (data[:, -2], data[:, -1]),
                                              self.hcol_indices[-1])

        if first and last:
            return np.hstack((np.expand_dims(first_column, 1),
                              data,
                              np.expand_dims(last_column, 1)))
        elif first:
            return np.hstack((np.expand_dims(first_column, 1),
                              data))
        elif last:
            return np.hstack((data,
                              np.expand_dims(last_column, 1)))
        else:
            return data

    def _fill_col_borders(self):
        """Add the first and last column to the data by extrapolation.
        """

        first = True
        last = True
        if self.col_indices[0] == self.hcol_indices[0]:
            first = False
        if self.col_indices[-1] == self.hcol_indices[-1]:
            last = False
        for num, data in enumerate(self.tie_data):
            self.tie_data[num] = self._extrapolate_cols(data, first, last)

        if first and last:
            self.col_indices = np.concatenate((np.array([self.hcol_indices[0]]),
                                               self.col_indices,
                                               np.array([self.hcol_indices[-1]])))
        elif first:
            self.col_indices = np.concatenate((np.array([self.hcol_indices[0]]),
                                               self.col_indices))
        elif last:
            self.col_indices = np.concatenate((self.col_indices,
                                               np.array([self.hcol_indices[-1]])))

    def _extrapolate_rows(self, data, row_indices, first_index, last_index):
        """Extrapolate the rows of data, to get the first and last together
        with the data.
        """

        pos = row_indices[:2]
        first_row = _linear_extrapolate(pos,
                                        (data[0, :], data[1, :]),
                                        first_index)
        pos = row_indices[-2:]
        last_row = _linear_extrapolate(pos,
                                       (data[-2, :], data[-1, :]),
                                       last_index)
        return np.vstack((np.expand_dims(first_row, 0),
                          data,
                          np.expand_dims(last_row, 0)))

    def _fill_row_borders(self):
        """Add the first and last rows to the data by extrapolation.
        """
        lines = len(self.hrow_indices)
        chunk_size = self.chunk_size or lines
        factor = len(self.hrow_indices) / len(self.row_indices)

        tmp_data = []
        for num in range(len(self.tie_data)):
            tmp_data.append([])
        row_indices = []

        for index in range(0, lines, chunk_size):
            indices = np.logical_and(self.row_indices >= index / factor,
                                     self.row_indices < (index
                                                         + chunk_size) / factor)
            ties = np.argwhere(indices).squeeze()
            tiepos = self.row_indices[indices].squeeze()

            for num, data in enumerate(self.tie_data):
                to_extrapolate = data[ties, :]
                if len(to_extrapolate) > 0:
                    extrapolated = self._extrapolate_rows(to_extrapolate,
                                                          tiepos,
                                                          self.hrow_indices[
                                                              index],
                                                          self.hrow_indices[index + chunk_size - 1])
                    tmp_data[num].append(extrapolated)

            row_indices.append(np.array([self.hrow_indices[index]]))
            row_indices.append(tiepos)
            row_indices.append(np.array([self.hrow_indices[index
                                                           + chunk_size - 1]]))

        for num in range(len(self.tie_data)):
            self.tie_data[num] = np.vstack(tmp_data[num])
        self.row_indices = np.concatenate(row_indices)

    def _interp(self):
        """Interpolate the cartesian coordinates.
        """
        if np.all(self.hrow_indices == self.row_indices):
            return self._interp1d()

        xpoints, ypoints = np.meshgrid(self.hrow_indices,
                                       self.hcol_indices)

        for num, data in enumerate(self.tie_data):
            spl = RectBivariateSpline(self.row_indices,
                                      self.col_indices,
                                      data,
                                      s=0,
                                      kx=self.kx_,
                                      ky=self.ky_)

            new_data_ = spl.ev(xpoints.ravel(), ypoints.ravel())
            self.new_data[num] = new_data_.reshape(xpoints.shape).T.copy(order='C')

    def _interp1d(self):
        """Interpolate in one dimension.
        """
        lines = len(self.hrow_indices)

        for num, data in enumerate(self.tie_data):
            self.new_data[num] = np.empty((len(self.hrow_indices),
                                           len(self.hcol_indices)),
                                          data.dtype)

            for cnt in range(lines):
                tck = splrep(self.col_indices, data[cnt, :], k=self.ky_, s=0)
                self.new_data[num][cnt, :] = splev(
                    self.hcol_indices, tck, der=0)

    def interpolate(self):
        """Do the interpolation, and return resulting longitudes and latitudes.
        """
        self._interp()

        return self.new_data
